"""Processing module for alarm actions."""

import logging
import time
from copy import copy
from datetime import datetime

import soco  # type: ignore
import soco.alarms  # type: ignore
import tabulate  # type: ignore
from soco.exceptions import SoCoUPnPException  # type: ignore

from soco_cli.utils import (
    convert_true_false,
    error_report,
    one_parameter,
    parameter_type_error,
    two_parameters,
    zero_parameters,
)

# Stabilisation delay (in seconds) when saving multiple alarms
ALARM_SAVE_DELAY = 1.0


@zero_parameters
def list_alarms(speaker, action, args, soco_function, use_local_speaker_list):
    alarms = soco.alarms.get_alarms(speaker)
    if not alarms:
        return True
    details = []
    for alarm in alarms:
        # Skip this alarm if it's not for the target room
        if action == "alarms_zone":
            if alarm.zone != speaker:
                continue
        didl = alarm.program_metadata
        title_start = didl.find("<dc:title>")
        if title_start >= 0:
            title_start += len("<dc:title>")
            title_end = didl.find("</dc:title>")
            title = didl[title_start:title_end]
        else:
            title = "Unknown"
        time = alarm.start_time.strftime("%H:%M")
        if alarm.duration:
            duration = alarm.duration.strftime("%H:%M")
        else:
            duration = "No Limit"
        details.append(
            [
                alarm.alarm_id,
                alarm.zone.player_name,
                time,
                duration,
                alarm.recurrence,
                convert_true_false(alarm.enabled),
                title,
                alarm.play_mode,
                alarm.volume,
                convert_true_false(alarm.include_linked_zones),
            ]
        )

    # Sort alarms by start time, room. Apply sorts in reverse order.
    details.sort(key=lambda field: field[1])  # Room
    details.sort(key=lambda field: field[2].split(":")[1])  # Minute
    details.sort(key=lambda field: field[2].split(":")[0])  # Hour

    headers = [
        "Alarm ID",
        "Speaker",
        "1: Start Time",
        "2: Duration",
        "3: Recurrence",
        "4: Enabled",
        "5: Title",
        "6: Play Mode",
        "7: Vol.",
        "8: Incl. Grouped",
    ]
    print()
    print(tabulate.tabulate(details, headers, tablefmt="github", numalign="left"))
    print()
    return True


@one_parameter
def remove_alarms(speaker, action, args, soco_function, use_local_speaker_list):

    alarms = soco.alarms.get_alarms(speaker)

    if args[0].lower() == "all":
        for alarm in alarms:
            logging.info("Removing alarm ID '{}'".format(alarm.alarm_id))
            alarm.remove()
        return True

    alarm_ids_to_delete = args[0].split(",")
    alarm_ids_to_delete = set(alarm_ids_to_delete)
    logging.info("Attempting to delete alarm ID(s): {}".format(alarm_ids_to_delete))

    alarm_ids = {alarm.alarm_id for alarm in alarms}
    logging.info("Current alarm ID(s): {}".format(alarm_ids))

    valid_alarm_ids_to_delete = alarm_ids.intersection(alarm_ids_to_delete)
    logging.info("Valid alarm ID(s) to delete: {}".format(valid_alarm_ids_to_delete))

    for alarm in alarms:
        if alarm.alarm_id in valid_alarm_ids_to_delete:
            logging.info("Deleting alarm ID: {}".format(alarm.alarm_id))
            alarm.remove()

    alarms_invalid = alarm_ids_to_delete.difference(valid_alarm_ids_to_delete)
    if len(alarms_invalid) != 0:
        print("Error: Alarm ID(s) {} not found".format(alarms_invalid))

    return True


@one_parameter
def add_alarm(speaker, action, args, soco_function, use_local_speaker_list):
    new_alarm = soco.alarms.Alarm(zone=speaker)
    if not _modify_alarm_object(new_alarm, args[0]):
        return False

    try:
        new_alarm.save()
    except soco.exceptions.SoCoUPnPException:
        error_report("Failed to create alarm")
        return False

    return True


@two_parameters
def modify_alarm(speaker, action, args, soco_function, use_local_speaker_list):

    alarm_ids = args[0].lower().split(",")
    all_alarms = soco.alarms.get_alarms(speaker)
    if alarm_ids[0] == "all":
        alarms = set(all_alarms)
    else:
        alarms = set()
        for alarm_id in alarm_ids:
            for alarm in all_alarms:
                if alarm_id == alarm.alarm_id:
                    alarms.add(alarm)
                    break
            else:
                print("Alarm ID '{}' not found".format(alarm_id))

    for index, alarm in enumerate(alarms):
        if not _modify_alarm_object(alarm, args[1]):
            continue
        try:
            logging.info("Saving alarm '{}'".format(alarm.alarm_id))
            alarm.save()
            if index < len(alarms) - 1:
                # Allow alarm update to stabilise
                logging.info(
                    "Waiting {}s after saving alarm '{}'".format(
                        ALARM_SAVE_DELAY, alarm.alarm_id
                    )
                )
                time.sleep(ALARM_SAVE_DELAY)
        except soco.exceptions.SoCoUPnPException:
            error_report("Failed to modify alarm {}".format(alarm.alarm_id))
            continue

    return True


@one_parameter
def copy_alarm(speaker, action, args, soco_function, use_local_speaker_list):
    """Copy an alarm to the target speaker."""
    return move_or_copy_alarm(speaker, args[0], copy=True)


@one_parameter
def move_alarm(speaker, action, args, soco_function, use_local_speaker_list):
    """Move an alarm to the target speaker."""
    return move_or_copy_alarm(speaker, args[0], copy=False)


def move_or_copy_alarm(speaker, alarm_id, copy=True):
    alarms = soco.alarms.get_alarms(speaker)
    for alarm in alarms:
        if alarm_id == alarm.alarm_id:
            break
    else:
        error_report("Alarm ID '{}' not found".format(alarm_id))
        return False

    if alarm.zone == speaker:
        error_report("Cannot copy/move an alarm to the same speaker")
        return False

    alarm.zone = speaker
    if copy is True:
        alarm._alarm_id = None
    try:
        alarm.save()
    except soco.exceptions.SoCoUPnPException:
        error_report("Failed to copy/move alarm")
        return False

    if copy is True:
        print("Alarm ID '{}' created".format(alarm.alarm_id))

    return True


@one_parameter
def enable_alarms(speaker, action, args, soco_function, use_local_speaker_list):
    return set_alarms(speaker, args[0], enabled=True)


@one_parameter
def disable_alarms(speaker, action, args, soco_function, use_local_speaker_list):
    return set_alarms(speaker, args[0], enabled=False)


def set_alarms(speaker, alarm_ids, enabled=True):
    alarms = soco.alarms.get_alarms(speaker)
    alarm_ids = set(alarm_ids.lower().split(","))

    all_alarms = False
    if "all" in alarm_ids:
        all_alarms = True
        alarm_ids.discard("all")
    elif "_all_" in alarm_ids:
        all_alarms = True
        alarm_ids.discard("_all_")

    for alarm in alarms:
        if all_alarms is True or alarm.alarm_id in alarm_ids:
            logging.info(
                "Setting alarm id '{}' to enabled = {}".format(alarm.alarm_id, enabled)
            )
            if alarm.enabled != enabled:
                alarm.enabled = enabled
                logging.info("Saving alarm '{}'".format(alarm.alarm_id))
                alarm.save()
                if len(alarm_ids) != 0 or all_alarms:
                    # Stabilisation delay
                    logging.info(
                        "Waiting {}s after saving alarm '{}'".format(
                            ALARM_SAVE_DELAY, alarm.alarm_id
                        )
                    )
                    time.sleep(ALARM_SAVE_DELAY)
            alarm_ids.discard(alarm.alarm_id)

    if len(alarm_ids) != 0:
        print("Alarm IDs not found: {}".format(alarm_ids))

    return True


@one_parameter
def snooze_alarm(speaker, action, args, soco_function, use_local_speaker_list):
    """Snooze an alarm that's currently playing"""

    duration = args[0].lower()

    # HH:MM:SS format
    h_m_s = duration.split(":")
    if len(h_m_s) == 3:
        try:
            if not (
                0 <= int(h_m_s[0]) <= 23
                and 0 <= int(h_m_s[1]) <= 59
                and 0 <= int(h_m_s[2]) <= 59
            ):
                raise ValueError
        except (ValueError, TypeError):
            logging.info("Invalid snooze duration: '{}'".format(args[0]))
            parameter_type_error(
                action,
                "A valid HH:MM:SS duration, or an integer number of minutes",
            )
            return False

    # Simple 'Nm' or 'N' for N minutes of snooze
    else:
        try:
            duration = abs(int(duration.replace("m", "")))
            minutes = str(duration % 60).zfill(2)
            hours = str(int(duration / 60)).zfill(2)
            duration = hours + ":" + minutes + ":00"
        except ValueError:
            logging.info("Invalid snooze duration: '{}'".format(args[0]))
            parameter_type_error(
                action,
                "An integer number of minutes, or HH:MM:SS format",
            )
            return False

    logging.info("Sending snooze command using duration '{}'".format(duration))
    try:
        speaker.avTransport.SnoozeAlarm([("InstanceID", 0), ("Duration", duration)])
    except SoCoUPnPException as error:
        logging.info("Exception: {}".format(error))
        if error.error_code == "701":
            error_report("Can only snooze a playing alarm")
        elif error.error_code == "402":
            error_report("Invalid snooze duration: '{}'".format(duration))
        else:
            error_report("{}".format(error))
        return False

    return True


@two_parameters
def copy_modify_alarm(speaker, action, args, soco_function, use_local_speaker_list):

    alarm_id = args[0]
    alarm_parms = args[1]

    # Find the alarm
    alarms = soco.alarms.get_alarms(speaker)
    for alarm in alarms:
        if alarm_id == alarm.alarm_id:
            break
    else:
        error_report(
            "Alarm ID '{}' not found; use the 'alarms' action to find the integer ID".format(
                alarm_id
            )
        )
        return False

    # Create a new alarm from the existing one
    new_alarm = copy(alarm)
    new_alarm._alarm_id = None
    new_alarm.zone = speaker

    # Apply modifications
    if not _modify_alarm_object(new_alarm, alarm_parms):
        return False

    # Save the new alarm
    try:
        new_alarm.save()
    except soco.exceptions.SoCoUPnPException:
        error_report("Failed to copy/move alarm; did you modify the start time?")
        return False

    return True


def _modify_alarm_object(alarm: soco.alarms.Alarm, parms_string: str) -> bool:

    alarm_parameters = parms_string.split(",")
    if len(alarm_parameters) != 8:
        error_report(
            "8 comma-separated parameters required for alarm modification specification"
        )
        return False

    start_time = alarm_parameters[0]
    if not start_time == "_":
        try:
            alarm.start_time = datetime.strptime(start_time, "%H:%M").time()
        except ValueError:
            error_report("Invalid time format: {}".format(start_time))
            return False

    duration = alarm_parameters[1]
    if not duration == "_":
        try:
            alarm.duration = datetime.strptime(duration, "%H:%M").time()
        except ValueError:
            error_report("Invalid time format: {}".format(duration))
            return False

    recurrence = alarm_parameters[2]
    if not recurrence == "_":
        if not soco.alarms.is_valid_recurrence(recurrence):
            error_report("'{}' is not a valid recurrence string".format(recurrence))
            return False
        alarm.recurrence = recurrence

    enabled = alarm_parameters[3].lower()
    if not enabled == "_":
        if enabled in ["on", "yes"]:
            enabled = True
        elif enabled in ["off", "no"]:
            enabled = False
        else:
            error_report(
                "Alarm must be enabled 'on' or 'off', not '{}'".format(
                    alarm_parameters[3]
                )
            )
            return False
        alarm.enabled = enabled

    uri = alarm_parameters[4]
    if not uri == "_":
        if uri.lower() == "chime":
            uri = None
        alarm.program_uri = uri

    play_mode = alarm_parameters[5].upper()
    if not play_mode == "_":
        play_mode_options = [
            "NORMAL",
            "SHUFFLE_NOREPEAT",
            "SHUFFLE",  # Note: this means SHUFFLE and REPEAT
            "REPEAT_ALL",
            "REPEAT_ONE",
            "SHUFFLE_REPEAT_ONE",
        ]
        if play_mode not in play_mode_options:
            error_report(
                "Play mode is '{}', should be one of:\n  {}".format(
                    alarm_parameters[5], play_mode_options
                )
            )
            return False
        alarm.play_mode = play_mode

    volume = alarm_parameters[6]
    if not volume == "_":
        try:
            volume = int(volume)
            if not 0 <= volume <= 100:
                error_report(
                    "Alarm volume must be between 0 and 100, not '{}'".format(
                        alarm_parameters[6]
                    )
                )
                return False
        except ValueError:
            error_report(
                "Alarm volume must be an integer between 0 and 100, not '{}'".format(
                    alarm_parameters[6]
                )
            )
            return False
        alarm.volume = volume

    include_linked = alarm_parameters[7].lower()
    if not include_linked == "_":
        if include_linked in ["on", "yes"]:
            include_linked = True
        elif include_linked in ["off", "no"]:
            include_linked = False
        else:
            error_report(
                "Linked zones must be enabled 'on' or 'off', not '{}'".format(
                    alarm_parameters[7]
                )
            )
            return False
        alarm.include_linked_zones = include_linked

    return True



"""Prints a table of information about the Sonos system."""

import datetime

import tabulate  # type: ignore

# Collect speaker information from each speaker in turn
headers = [
    "Zone Name",
    "IP Address",
    "Visible",
    "CoOrd",
    "CoOrd IP",
    "Vol.",
    "Mute",
    "State",
    "Model Name",
    "Model No.",
    "HW Version",
    "SW Version",
]


def print_speaker_table(device):
    speakers = []
    models = set()
    errors = []
    exceptions = []

    def add_err_and_exc(err_player_name, err_ip_address, exc):
        exceptions.append(exc)
        errors.append(
            ("Could not get speaker_info for {}: {}").format(
                err_player_name, err_ip_address
            )
        )

    for sco in device.all_zones:
        # Load the speaker info
        try:
            sco.get_speaker_info()
        except BaseException as e:
            add_err_and_exc(sco.player_name, sco.ip_address, e)
            continue

        # Boost and Bridge don't support some attributes
        if sco.is_bridge:
            not_applicable = "n/a"
            volume = not_applicable
            mute = not_applicable
            state = not_applicable
        else:
            volume = sco.volume
            mute = "On" if sco.mute else "Off"
            # Bonded speakers return errors for transport and track info.
            # Wrap in an exception, and ignore.
            try:
                state = sco.get_current_transport_info()["current_transport_state"]
            except BaseException as e:
                # If we're here, assume the speakers are bonded
                # in a Home Theatre configuration
                state = "Bonded"

        # Find the coordinator IP
        coord = sco.group.coordinator
        if sco is coord:
            coord_ip = ""
        else:
            coord_ip = coord.ip_address
            # Overwrite the 'state'
            if state != "Bonded":
                state = sco.group.coordinator.get_current_transport_info()[
                    "current_transport_state"
                ]

        # Set up the information for the speaker
        speaker = [
            sco.player_name,
            sco.ip_address,
            "Yes" if sco.is_visible else "No",
            "Yes" if sco.is_coordinator else "No",
            coord_ip,
            volume,
            mute,
            state,
            sco.speaker_info["model_name"].replace("Sonos ", ""),
            sco.speaker_info["model_number"],
            sco.speaker_info["hardware_version"],
            sco.speaker_info["software_version"]
            + " ("
            + sco.speaker_info["display_version"]
            + ")",
        ]
        speakers.append(speaker)
        models.add(sco.speaker_info["model_number"])

    # Print the date and time
    print()
    print(
        "Report generated on:",
        datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M UTC%z (%A)"),
    )

    # Print the speaker information table in a nice format
    print()
    print(tabulate.tabulate(sorted(speakers), headers, numalign="center"))

    # Print the list of unique model numbers
    print("\nSonos model numbers present:", end="")
    for index, model in enumerate(sorted(models)):
        print(" " + model, end="")
        if index != len(models) - 1:
            print(",", end="")
        else:
            print(".", end="")
    print(
        "\nDevice counts: {} total Sonos device(s), {} unique model(s).".format(
            len(speakers), len(models)
        )
    )
    print()

    # List any speakers that couldn't be inspected, along with the
    # relevant exception
    if len(errors) != 0:
        for err, exc in zip(errors, exceptions):
            print(err)
            print("Exception: {}".format(exc))
        print()
        return False
    return True



"""Plays a list of files from the local filesystem, with interactive options."""

import logging
import os
import sys
from multiprocessing import Process
from os import chdir, name, path, scandir
from pathlib import Path
from random import choice, sample
from typing import List

from soco import SoCo  # type: ignore

from soco_cli.m3u_parser import parse_m3u
from soco_cli.play_local_file import is_supported_type, play_local_file
from soco_cli.utils import error_report


def interaction_manager(speaker_ip: str) -> None:
    sys.stdin = open(0)
    speaker = SoCo(speaker_ip)
    while True:
        try:
            # keypress = wait_for_keypress()
            keypress = input("")[0]
        except:
            keypress = ""
        if keypress in ["N", "n"]:
            action = "NEXT"
            print("Next track ...")
            speaker.stop()
            logging.info(
                "Interactive mode: key = '{}', action = '{}'".format(keypress, action)
            )
        if keypress in ["P", "p"]:
            action = "PAUSE"
            print("Pause playback ...")
            try:
                speaker.pause()
            except Exception as e:
                logging.info("Exception ignored: {}".format(e))
            logging.info(
                "Interactive mode: key = '{}', action = '{}'".format(keypress, action)
            )
        if keypress in ["R", "r"]:
            action = "RESUME"
            print("Resume playback ...")
            try:
                speaker.play()
            except Exception as e:
                logging.info("Exception ignored: {}".format(e))
            logging.info(
                "Interactive mode: key = '{}', action = '{}'".format(keypress, action)
            )
        # Windows captures CTRL-C key-presses, so we handle them directly here
        if name == "nt" and keypress == "\x03":
            logging.info(
                "Windows CTRL-C: Stopping speaker '{}' and exiting".format(
                    speaker.player_name
                )
            )
            speaker.stop()
            os._exit(0)


def play_file_list(speaker: SoCo, tracks: List[str], options: str = "") -> bool:
    """Play a list of files (tracks) with absolute pathnames."""
    options = options.lower()

    # Check for invalid options
    invalid = set(options) - set("psri")
    if invalid:
        error_report("Invalid option(s) '{}' supplied".format(invalid))
        return False

    if options != "":
        # Grab back stdout from api.run_command()
        sys.stdout = sys.__stdout__

    if "r" in options:
        # Choose a single random track
        track = choice(tracks)
        tracks = [track]
        logging.info("Choosing random track: {}".format(track))

    elif "s" in options:
        logging.info("Shuffling playlist")
        # For some reason, 'shuffle(tracks)' does not work
        tracks = sample(tracks, len(tracks))

    # Interactive mode
    keypress_process = None
    if "i" in options:
        print("Interactive mode actions: (N)ext, (P)ause, (R)esume + RETURN")
        try:
            logging.info("Interactive mode ... starting keypress process")
            keypress_process = Process(
                target=interaction_manager, args=(speaker.ip_address,), daemon=True
            )
            keypress_process.start()
            logging.info("Process PID {} created".format(keypress_process.pid))
        except Exception as e:
            logging.info("Exception ignored: {}".format(e))
            keypress_process = None

    zero_pad = len(str(len(tracks)))
    for index, track in enumerate(tracks):

        if not path.exists(track):
            print("Error: file not found:", track)
            continue

        if not is_supported_type(track):
            print("Error: unsupported file type:", track)
            continue

        if "p" in options:
            print(
                "Playing {} of {}:".format(str(index + 1).zfill(zero_pad), len(tracks)),
                track,
            )

        play_local_file(speaker, track)

    if keypress_process:
        keypress_process.terminate()

    return True


def play_m3u_file(speaker: SoCo, m3u_file: str, options: str = "") -> bool:
    if not path.exists(m3u_file):
        error_report("File '{}' not found".format(m3u_file))
        return False

    logging.info("Parsing file contents'{}'".format(m3u_file))
    track_list = parse_m3u(m3u_file)
    if len(track_list) == 0:
        error_report("No tracks found in '{}'".format(m3u_file))
        return False

    directory, _ = path.split(m3u_file)
    if directory != "":
        chdir(directory)
    tracks = [str(Path(track.path).absolute()) for track in track_list]  # type:ignore
    logging.info("Files to to play: {}".format(tracks))

    play_file_list(speaker, tracks, options)
    return True


def play_directory_files(speaker: SoCo, directory: str, options: str = "") -> bool:
    """Play all the valid audio files in a directory. Ignores subdirectories"""
    tracks = []
    try:
        with scandir(directory) as files:
            for file in files:
                if is_supported_type(file.name):
                    tracks.append(path.abspath(path.join(directory, file.name)))
    except FileNotFoundError:
        error_report("Directory '{}' not found".format(directory))
        return False

    tracks.sort()
    logging.info("Files to to play: {}".format(tracks))
    play_file_list(speaker, tracks, options)
    return True



"""Implements an HTTP API server for SoCo-CLI commands."""

from sys import version_info

if version_info.major == 3 and version_info.minor < 6:
    print("HTTP API Server requires Python 3.6 or above")
    exit(1)

import argparse
import shlex
from os.path import abspath
from subprocess import check_output
from typing import Dict

import uvicorn  # type: ignore
from fastapi import FastAPI

from soco_cli.__init__ import __version__ as version  # type: ignore
from soco_cli.api import get_all_speaker_names
from soco_cli.api import get_soco_object as get_speaker
from soco_cli.api import rescan_speakers
from soco_cli.api import run_command as sc_run
from soco_cli.utils import version as print_version

# Globals
USE_LOCAL = False
PORT = 8000
INFO = "SoCo-CLI HTTP API Server v" + version
PREFIX = "SoCo-CLI: "
MACROS: Dict[str, str] = {}
MACRO_FILE = ""


sc_app = FastAPI()


def command_core(
    speaker: str, action: str, *args: str, use_local: bool = False
) -> Dict:
    device, error_msg = get_speaker(speaker, use_local_speaker_list=use_local)
    if device:
        speaker = device.player_name
        exit_code, result, error_msg = sc_run(
            device, action, *args, use_local_speaker_list=use_local
        )
    else:
        exit_code = 1
        result = ""

    # Quote speaker names & arguments containing spaces, for neatness
    if " " in speaker:
        quoted_speaker = '"' + speaker + '"'
    else:
        quoted_speaker = speaker
    new_args = []
    for i in range(len(args)):
        if " " in args[i]:
            new_args.append('"' + args[i] + '"')
        else:
            new_args.append(args[i])

    # Print the equivalent 'sonos' command and exit code
    if len(new_args) != 0:
        arguments = " ".join(new_args).rstrip()
        print(
            PREFIX
            + "Command = 'sonos {} {} {}', ".format(quoted_speaker, action, arguments),
            end="",
        )
    else:
        print(
            PREFIX + "Command = 'sonos {} {}', ".format(quoted_speaker, action), end=""
        )
    if exit_code == 0:
        print("exit code = {}".format(exit_code))
    else:
        print("exit code = {} [{}]".format(exit_code, error_msg))

    return {
        "speaker": speaker,
        "action": action,
        "args": args,
        "exit_code": exit_code,
        "result": result,
        "error_msg": error_msg,
    }


@sc_app.get("/")
async def root() -> Dict:
    return {"info": INFO}


@sc_app.get("/rediscover")
async def rediscover() -> Dict:
    rescan_speakers(timeout=2.0)
    speakers = get_all_speaker_names()
    print(PREFIX + "Speakers (re)discovered: {}".format(speakers))
    return {"speakers_discovered": speakers}


@sc_app.get("/speakers")
async def speakers() -> Dict:
    speakers = get_all_speaker_names()
    print(PREFIX + "Speakers: {}".format(speakers))
    return {"speakers": speakers}


@sc_app.get("/macro/{macro_name}")
async def run_commands(macro_name: str) -> Dict:
    result = _process_macro(macro_name)
    return {"result": result}


@sc_app.get("/{speaker}/{action}")
async def action_0(speaker: str, action: str) -> Dict:
    return command_core(speaker, action, use_local=USE_LOCAL)


@sc_app.get("/{speaker}/{action}/{arg_1}")
async def action_1(speaker: str, action: str, arg_1: str) -> Dict:
    return command_core(speaker, action, arg_1, use_local=USE_LOCAL)


@sc_app.get("/{speaker}/{action}/{arg_1}/{arg_2}")
async def action_2(speaker: str, action: str, arg_1: str, arg_2: str) -> Dict:
    return command_core(speaker, action, arg_1, arg_2, use_local=USE_LOCAL)


@sc_app.get("/{speaker}/{action}/{arg_1}/{arg_2}/{arg_3}")
async def action_3(
    speaker: str, action: str, arg_1: str, arg_2: str, arg_3: str
) -> Dict:
    return command_core(speaker, action, arg_1, arg_2, arg_3, use_local=USE_LOCAL)


def args_processor() -> None:
    parser = argparse.ArgumentParser(
        prog="sonos-http-api-server",
        usage="%(prog)s",
        description=INFO,
    )
    parser.add_argument(
        "--port",
        "-p",
        type=int,
        help="The port on which to listen",
    )
    parser.add_argument(
        "--version",
        "-v",
        action="store_true",
        default=False,
        help="Print the SoCo-CLI and SoCo versions, and exit",
    )

    parser.add_argument(
        "--macros",
        "-m",
        type=str,
        default="macros.txt",
        help="The file containing the local macros",
    )

    args = parser.parse_args()

    if args.version:
        print_version()
        exit(0)

    global PORT
    if args.port is not None:
        PORT = args.port

    global MACRO_FILE
    MACRO_FILE = abspath(args.macros)


def main() -> None:
    args_processor()
    print(PREFIX + "Starting " + INFO)

    # Load local macros
    global MACROS
    _load_macros(MACROS, filename=MACRO_FILE)

    try:
        # Pre-load speaker cache
        print(PREFIX + "Finding speakers ... ", end="", flush=True)
        try:
            # This forces speaker discovery
            # For some reason, using 'get_all_speakers()' generates Uvicorn errors
            get_speaker("")
            print(get_all_speaker_names())
        except:
            print("discovery failed, will retry on first request")

        # Start the server
        uvicorn.run(sc_app, host="0.0.0.0", use_colors=False, port=PORT)
        print(PREFIX + INFO + " stopped")
        exit(0)

    except Exception as error:
        print("Error: {}".format(error))
        exit(1)


def _process_macro(macro_name: str) -> str:
    # Look up the macro
    try:
        macro = _lookup_macro(macro_name)
    except KeyError:
        print(PREFIX + "macro '{}' not found".format(macro_name))
        return "Error: macro '{}' not found".format(macro_name)

    # Substitute speaker names for IP addresses, for efficiency
    sonos_command_line = "sonos " + _substitute_speaker_ips(macro)
    print(PREFIX + "Executing: " + sonos_command_line)

    # Execute the command
    try:
        output = check_output(sonos_command_line, shell=True)
        return "Command line output: " + output.decode("utf-8")
    except:
        return "Error running command line: " + sonos_command_line


def _lookup_macro(macro_name: str) -> str:
    global MACROS
    return MACROS[macro_name]


def _substitute_speaker_ips(macro: str, use_local: bool = False) -> str:
    """
    Substitute speaker names for IP addresses, for efficiency.
    Speaker names must be exact.
    """
    terms = shlex.split(macro)
    new_macro_list = []
    for term in terms:
        device, error_msg = get_speaker(term, use_local_speaker_list=use_local)
        if device is not None and device.player_name == term:
            new_macro_list.append(device.ip_address)
        else:
            new_macro_list.append(term)
    new_macro = (" ").join(new_macro_list)
    return new_macro


def _load_macros(macros: dict, filename: str) -> bool:
    print(PREFIX + "Attempting to load macros from '{}'".format(filename))
    try:
        with open(filename, "r") as f:
            line = f.readline()
            while line != "":
                if not line.startswith("#") and line != "\n":
                    if line.count("=") != 1:
                        print(PREFIX + "Malformed macro '{}'... ignored".format(line))
                        print(line, end="")
                    else:
                        macro = line.split("=")
                        macros[macro[0].strip()] = macro[1].strip()
                line = f.readline()
        print(PREFIX + "Loaded macros: {}".format(macros))
        return True
    except:
        print(PREFIX + "Macro file not found")
        return False


if __name__ == "__main__":
    main()



"""Common utilities used across multiple modules."""

import datetime
import logging
import os
import pickle
import signal

try:
    import readline
except ImportError:
    pass
import sys
from collections.abc import Sequence
from platform import python_version
from time import sleep

import soco  # type: ignore

from soco_cli.__init__ import __version__  # type: ignore
from soco_cli.match_speaker_names import speaker_name_matches
from soco_cli.speakers import Speakers


def event_unsubscribe(sub):
    """Unsubscribe from events, with a try/catch wrapper, and a pause
    introduced to yield the thread."""

    logging.info("Unsubscribing '{}'".format(sub))
    try:
        sleep(0.2)
        sub.unsubscribe()
    except Exception as e:
        logging.info("Failed to unsubscribe: {}".format(e))
    logging.info("Unsubscribed")


INTERACTIVE = False
API = False
SINGLE_KEYSTROKE = False


def set_interactive():
    global INTERACTIVE
    INTERACTIVE = True


def set_api():
    global API
    API = True


def set_single_keystroke(sk):
    global SINGLE_KEYSTROKE
    SINGLE_KEYSTROKE = sk


# Error handling
def error_report(msg):
    # Print to stderr
    print("Error:", msg, file=sys.stderr, flush=True)
    # Use os._exit() to avoid the catch-all 'except'
    if not (INTERACTIVE or API):
        logging.info("Exiting program using os._exit(1)")
        os._exit(1)


def parameter_type_error(action, required_params):
    msg = "Action '{}' takes parameter(s): {}".format(action, required_params)
    error_report(msg)


def parameter_number_error(action, parameter_number):
    msg = "Action '{}' takes {} parameter(s)".format(action, parameter_number)
    error_report(msg)


# Parameter count checking
def zero_parameters(f):
    def wrapper(*args, **kwargs):
        if len(args[2]) != 0:
            parameter_number_error(args[1], "no")
            return False

        return f(*args, **kwargs)

    return wrapper


def one_parameter(f):
    def wrapper(*args, **kwargs):
        if len(args[2]) != 1:
            parameter_number_error(args[1], "1")
            return False
        return f(*args, **kwargs)

    return wrapper


def zero_or_one_parameter(f):
    def wrapper(*args, **kwargs):
        if len(args[2]) not in [0, 1]:
            parameter_number_error(args[1], "0 or 1")
            return False
        return f(*args, **kwargs)

    return wrapper


def one_or_two_parameters(f):
    def wrapper(*args, **kwargs):
        if len(args[2]) not in [1, 2]:
            parameter_number_error(args[1], "1 or 2")
            return False
        return f(*args, **kwargs)

    return wrapper


def two_parameters(f):
    def wrapper(*args, **kwargs):
        if len(args[2]) != 2:
            parameter_number_error(args[1], "2")
            return False
        return f(*args, **kwargs)

    return wrapper


def zero_one_or_two_parameters(f):
    def wrapper(*args, **kwargs):
        if len(args[2]) > 2:
            parameter_number_error(args[1], "zero, one or two")
            return False
        return f(*args, **kwargs)

    return wrapper


def one_or_more_parameters(f):
    def wrapper(*args, **kwargs):
        if len(args[2]) < 1:
            parameter_number_error(args[1], "1 or more")
            return False
        return f(*args, **kwargs)

    return wrapper


# Time manipulation
def seconds_until(time_str):
    # target_time = datetime.time.fromisoformat(time_str)
    target_time = create_time_from_str(time_str)
    now_time = datetime.datetime.now().time()
    delta_target = datetime.timedelta(
        hours=target_time.hour, minutes=target_time.minute, seconds=target_time.second
    )
    delta_now = datetime.timedelta(
        hours=now_time.hour, minutes=now_time.minute, seconds=now_time.second
    )
    diff = int((delta_target - delta_now).total_seconds())
    # Ensure 'past' times are treated as future times by adding 24hr
    return diff if diff > 0 else diff + 24 * 60 * 60


def create_time_from_str(time_str):
    """Process times in HH:MM(:SS) format. Return a 'time' object."""
    if ":" not in time_str:
        raise ValueError
    parts = time_str.split(":")
    if len(parts) not in [2, 3]:
        raise ValueError
    hours = int(parts[0])
    minutes = int(parts[1])
    if len(parts) == 3:
        seconds = int(parts[2])
    else:
        seconds = 0
    # Accept time strings from 00:00:00 to 23:59:59
    if 0 <= hours <= 23 and 0 <= minutes <= 59 and 0 <= seconds <= 59:
        return datetime.time(hour=hours, minute=minutes, second=seconds)

    raise ValueError


def convert_to_seconds(time_str):
    """Convert a time string to seconds.
    time_str can be one of Nh, Nm or Ns, or of the form HH:MM:SS
    :raises ValueError
    """
    logging.info("Converting '{}' to a number of seconds".format(time_str))
    time_str = time_str.lower()
    try:
        if ":" in time_str:  # Assume form is HH:MM:SS or HH:MM
            parts = time_str.split(":")
            if len(parts) == 3:  # HH:MM:SS
                td = datetime.timedelta(
                    hours=int(parts[0]), minutes=int(parts[1]), seconds=int(parts[2])
                )
            else:  # HH:MM
                td = datetime.timedelta(hours=int(parts[0]), minutes=int(parts[1]))
            return td.seconds

        if time_str.endswith("s"):  # Seconds (explicit)
            duration = float(time_str[:-1])
        elif time_str.endswith("m"):  # Minutes
            duration = float(time_str[:-1]) * 60
        elif time_str.endswith("h"):  # Hours
            duration = float(time_str[:-1]) * 60 * 60
        else:  # Seconds (default)
            duration = float(time_str)
        return duration
    except:
        raise ValueError


# Miscellaneous
def convert_true_false(true_or_false, conversion="YesOrNo"):
    if conversion == "YesOrNo":
        return "Yes" if true_or_false is True else "No"
    if conversion == "onoroff":
        return "on" if true_or_false is True else "off"
    return None


def version():
    print("soco-cli version:   {}".format(__version__), flush=True)
    print("soco version:       {}".format(soco.__version__), flush=True)
    print("python version:     {}".format(python_version()), flush=True)
    print("command path:       {}".format(sys.argv[0]), flush=True)


def docs():
    version = "v{}".format(__version__)
    if __version__.endswith("+"):
        url = "https://github.com/avantrec/soco-cli/blob/next_version/README.md"
    else:
        url = "https://github.com/avantrec/soco-cli/blob/{}/README.md".format(version)
    print("Online documentation for {}: {}".format(version, url), flush=True)


def logo():
    version = "v{}".format(__version__)
    if __version__.endswith("+"):
        url = "https://raw.githubusercontent.com/avantrec/soco-cli/next_version/assets/soco-cli-logo-01-large.png"
    else:
        url = "https://raw.githubusercontent.com/avantrec/soco-cli/{}/assets/soco-cli-logo-01-large.png".format(
            version
        )
    print("SoCo-CLI Logo: {}".format(url), flush=True)


# Suspend signal handling processing for 'exec' in interactive shell
suspend_sighandling = False


def set_suspend_sighandling(suspend=True):
    global suspend_sighandling
    logging.info("Setting 'suspend_sighandling' to '{}'".format(suspend))
    suspend_sighandling = suspend


# Stop a stream if playing a local file
speaker_playing_local_file = None


def set_speaker_playing_local_file(speaker):
    global speaker_playing_local_file
    if speaker:
        logging.info(
            "Setting speaker playing local file to '{}'".format(speaker.player_name)
        )
    else:
        logging.info("No speaker playing local file")
    speaker_playing_local_file = speaker


def sig_handler(signal_received, frame):

    logging.info("Caught signal: {}".format(signal_received))

    if suspend_sighandling:
        logging.info("Signal handling suspended ... ignoring")
        return

    # Restore stdout and stderr ... these have been redirected if
    # api.run_command() was used
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__

    # Prevent SIGINT (CTRL-C) exit: untidy exit from readline can leave
    # some terminals in a broken state
    if signal_received == signal.SIGINT:
        if SINGLE_KEYSTROKE:
            logging.info("SINGLE_KEYSTROKE set ... preventing exit")
            print("\nPlease use 'x' to exit >> ", end="", flush=True)
            return

        if INTERACTIVE:
            logging.info("INTERACTIVE set ... preventing exit")
            print("\nPlease use 'exit' to terminate the shell > ", end="", flush=True)
            if os.name == "nt":
                print(flush=True)
            return

    # Allow SIGTERM termination, but issue warning if interactive
    if signal_received == signal.SIGTERM and INTERACTIVE:
        print("\nSoCo-CLI process terminating ...", flush=True)
        print(
            "This can leave some terminals in a misconfigured state.",
            flush=True,
        )

    if speaker_playing_local_file:
        logging.info(
            "Speaker '{}': 'play_file' active ... stopping".format(
                speaker_playing_local_file.player_name
            )
        )
        speaker_playing_local_file.stop()

    logging.info("Unsubscribing from event notifications")
    unsub_all_remembered_event_subs()

    logging.info("Exiting program using 'os._exit(0)'")
    print("", flush=True)
    os._exit(0)


class RewindableList(Sequence):
    """This is a just-enough-implementation class to provide a list
    that can be rewound during iteration.
    """

    def __init__(self, items=[]):
        self._items = items
        self._index = 0

    def __iter__(self):
        self.rewind()
        return self

    def __getitem__(self, item):
        return self._items[item]

    def __len__(self):
        return len(self._items)

    def __next__(self):
        if self._index < len(self._items):
            item = self._items[self._index]
            self._index += 1
            return item
        raise StopIteration

    def rewind(self):
        self._index = 0

    def rewind_to(self, index):
        if len(self._items) == 0 and index == 0:
            self._index = 0
        elif 0 <= index < len(self._items):
            self._index = index
        else:
            raise IndexError

    def __str__(self):
        return str(self._items)

    def index(self):
        return self._index

    def insert(self, index, element):
        self._items.insert(index, element)
        if index <= self._index:
            self._index += 1

    def pop_next(self):
        item = self._items.pop(0)
        if self._index != 0:
            self._index -= 1
        return item


# Set up logging
def configure_logging(log_level: str) -> None:
    log_level = log_level.lower()
    if log_level == "none":
        # Disables all logging (i.e., CRITICAL and below)
        logging.disable(logging.CRITICAL)
    else:
        log_format = (
            "%(asctime)s %(filename)s:%(lineno)s - %(funcName)s() - %(message)s"
        )
        if log_level == "debug":
            logging.basicConfig(format=log_format, level=logging.DEBUG)
        elif log_level == "info":
            logging.basicConfig(format=log_format, level=logging.INFO)
        elif log_level in ["warn", "warning"]:
            logging.basicConfig(format=log_format, level=logging.WARNING)
        elif log_level == "error":
            logging.basicConfig(format=log_format, level=logging.ERROR)
        elif log_level == "critical":
            logging.basicConfig(format=log_format, level=logging.CRITICAL)
        else:
            error_report(
                "--log takes one of: NONE, DEBUG, INFO, WARN(ING), ERROR, CRITICAL"
            )


# Local speaker list operations
speaker_list = None


def set_speaker_list(s):
    global speaker_list
    speaker_list = s


class SpeakerCache:
    def __init__(self, max_threads=256, scan_timeout=0.1, min_netmask=24):
        # _cache contains (soco_instance, speaker_name) tuples
        self._cache = set()
        self._scan_done = False
        self._discovery_done = False
        self._max_threads = max_threads
        self._scan_timeout = scan_timeout
        self._min_netmask = min_netmask

    @property
    def exists(self):
        return bool(self._cache)

    def cache_speakers(self, speakers):
        logging.info("Adding speakers to cache: {}".format(speakers))
        for speaker in speakers:
            self._cache.add((speaker, speaker.player_name))

    def discover(self, reset=False):
        if not self._discovery_done or reset:
            # Clear the current cache
            self._cache = set()
            speakers = soco.discovery.discover(
                allow_network_scan=True,
                max_threads=self._max_threads,
                scan_timeout=self._scan_timeout,
                min_netmask=self._min_netmask,
            )
            if speakers:
                self.cache_speakers(speakers)
            else:
                logging.info("No speakers found to cache")
            self._discovery_done = True

    def scan(self, reset=False, scan_timeout_override=None):
        if not self._scan_done or reset:
            # Clear the current cache
            self._cache = set()
            scan_timeout = (
                scan_timeout_override if scan_timeout_override else self._scan_timeout
            )
            logging.info(
                "Performing full discovery scan with timeout = {}s".format(scan_timeout)
            )
            speakers = soco.discovery.scan_network(
                multi_household=True,
                max_threads=self._max_threads,
                scan_timeout=scan_timeout,
                min_netmask=self._min_netmask,
            )
            if speakers:
                self.cache_speakers(speakers)
                self._scan_done = True
            else:
                logging.info("No speakers found to cache")
        else:
            logging.info("Full discovery scan already done, and reset not requested")

    def add(self, speaker):
        logging.info("Adding speaker to cache")
        self._cache.add((speaker, speaker.player_name))

    def find_indirect(self, name):
        speakers_found = set()
        speakers_found_names = set()
        for cached, _ in self._cache:
            for speaker in cached.visible_zones:
                match, exact = speaker_name_matches(name, speaker.player_name)
                if match and exact:
                    return speaker
                if match and not exact:
                    speakers_found.add(speaker)
                    speakers_found_names.add(speaker.player_name)

        if len(speakers_found) == 1:
            return speakers_found.pop()

        if len(speakers_found) > 1:
            error_report("'{}' is ambiguous: {}".format(name, speakers_found_names))

        return None

    def find(self, name):
        speakers_found = set()
        speakers_found_names = set()
        for speaker, speaker_name in self._cache:
            match, exact = speaker_name_matches(name, speaker_name)
            if match and exact:
                return speaker
            if match and not exact:
                speakers_found.add(speaker)
                speakers_found_names.add(speaker_name)

        if len(speakers_found) == 1:
            return speakers_found.pop()

        if len(speakers_found) > 1:
            error_report(
                "Speaker name '{}' is ambiguous within {}".format(
                    name, speakers_found_names
                )
            )

        return None

    def get_all_speakers(self, use_scan=False):
        if use_scan:
            self.scan()
        else:
            self.discover()
        return self._cache

    def get_all_speaker_names(self, use_scan=False):
        if use_scan:
            self.scan()
        else:
            self.discover()
        names = [speaker[1] for speaker in self._cache]
        names.sort()
        return names

    def rename_speaker(self, old_name, new_name):
        for speaker in self._cache:
            if speaker[1] == old_name:
                logging.info("Updating speaker cache with new name")
                self._cache.remove(speaker)
                self._cache.add((speaker[0], new_name))
                return True
        logging.info("Speaker with name '{}' not found".format(old_name))
        return False


SPKR_CACHE = None


# Single instance of the speaker cache
def create_speaker_cache(max_threads=256, scan_timeout=1.0, min_netmask=24):
    global SPKR_CACHE
    SPKR_CACHE = SpeakerCache(
        max_threads=max_threads, scan_timeout=scan_timeout, min_netmask=min_netmask
    )


def speaker_cache():
    """Return the global speaker cache object"""
    return SPKR_CACHE


def local_speaker_list():
    """Return the global speaker list object"""
    return speaker_list


def get_speaker(name, local=False):
    # Use an IP address
    # (Allow the use of an IP address even if 'local' is specified)
    if Speakers.is_ipv4_address(name):
        logging.info("Using IP address instead of speaker name")
        return soco.SoCo(name)

    # Use the local speaker list
    if local:
        logging.info("Using local speaker list")
        return speaker_list.find(name)

    # Use discovery
    # Try various lookup methods in order of expense,
    # and cache results where possible
    speaker = None
    if not speaker:
        logging.info("Trying direct cache lookup")
        speaker = SPKR_CACHE.find(name)
    if not speaker:
        logging.info("Trying indirect cache lookup")
        speaker = SPKR_CACHE.find_indirect(name)
    if not speaker:
        logging.info("Trying standard discovery with network scan fallback")
        SPKR_CACHE.discover()
        speaker = SPKR_CACHE.find(name)
    if not speaker:
        logging.info("Trying network scan discovery")
        SPKR_CACHE.scan()
        speaker = SPKR_CACHE.find(name)
    if speaker:
        logging.info("Successful speaker discovery")
    else:
        logging.info("Failed to discover speaker")
    return speaker


def get_right_hand_speaker(left_hand_speaker):
    # Get the right-hand speaker of a stereo pair when the
    # left-hand speaker is supplied
    if not left_hand_speaker.is_visible:
        # If not visible, this is not a left-hand speaker
        logging.info("Speaker is visible: not a left-hand speaker")
        return None

    # Find the speaker which is not visible, for which the
    # left-hand speaker is the coordinator, and not a Sub
    for rh_speaker in left_hand_speaker.all_zones:
        if (
            rh_speaker.group.coordinator.ip_address == left_hand_speaker.ip_address
            and not rh_speaker.is_visible
            and "sub" not in rh_speaker.get_speaker_info()["model_name"].lower()
        ):
            logging.info(
                "Found right-hand speaker: {} / {}".format(
                    rh_speaker.player_name, rh_speaker.ip_address
                )
            )
            return rh_speaker
    logging.info("Right-hand speaker not found")
    return None


def rename_speaker_in_cache(old_name, new_name, use_local_speaker_list=True):
    if use_local_speaker_list:
        return speaker_list.rename(old_name, new_name)
    return SPKR_CACHE.rename_speaker(old_name, new_name)


# Argument processing
def configure_common_args(parser):
    """Set up the optional arguments common across the command line programs"""
    parser.add_argument(
        "--network-discovery-threads",
        "-t",
        type=int,
        default=256,
        help="Maximum number of threads for Sonos network discovery",
    )
    parser.add_argument(
        "--network-discovery-timeout",
        "-n",
        type=float,
        default=1.0,
        help="Network timeout for Sonos device scan (seconds)",
    )
    parser.add_argument(
        "--min_netmask",
        "-m",
        type=int,
        default=24,
        help="Minimum netmask for Sonos device scan (integer 0-32)",
    )
    parser.add_argument(
        "--version",
        "-v",
        action="store_true",
        default=False,
        help="Print the SoCo-CLI and SoCo versions and exit",
    )
    parser.add_argument(
        "--log",
        type=str,
        default="NONE",
        help="Set the logging level: 'NONE' (default) |'CRITICAL' | 'ERROR' | 'WARN'| 'INFO' | 'DEBUG'",
    )
    parser.add_argument(
        "--docs",
        action="store_true",
        default=False,
        help="Print the URL to the online documentation",
    )
    parser.add_argument(
        "--logo",
        action="store_true",
        default=False,
        help="Print the URL to the SoCo-CLI logo",
    )
    parser.add_argument(
        "--check_for_update",
        action="store_true",
        default=False,
        help="Check for a more recent version of SoCo-CLI",
    )


def check_args(args):
    """Check values of parameters. Returns None, or an error message."""
    message = ""
    if not 0 <= args.min_netmask <= 32:
        message = (
            message + "\n    Option 'min_netmask' must be an integer between 0 and 32"
        )
    if not 0.0 <= args.network_discovery_timeout <= 60.0:
        message = message + "\n    Option 'network_timeout' must be between 0.0 and 60s"
    if not 1 <= args.network_discovery_threads <= 32000:
        message = message + "\n    Option 'threads' must be between 1 and 32000"
    if message == "":
        return None
    return message


path = os.path.expanduser("~") + "/.soco-cli/"
filename = "saved_search.pickle"
pathname = path + filename


def save_search(result):
    if not os.path.exists(path):
        os.mkdir(path)
    with open(pathname, "wb") as f:
        pickle.dump(result, f)
    logging.info("Saved search results at {}".format(pathname))
    return True


def read_search():
    if os.path.exists(pathname):
        logging.info("Loading search results from {}".format(pathname))
        try:
            with open(pathname, "rb") as f:
                return pickle.load(f)
        except Exception as e:
            logging.info("Failed to load search results: %s", e)
    return None


# Interactive shell history file
SOCO_CLI_DIR = os.path.join(os.path.expanduser("~"), ".soco-cli")
HIST_FILE = os.path.join(SOCO_CLI_DIR, "shell-history.txt")
HIST_LEN = 50


def save_readline_history():
    if not os.path.exists(SOCO_CLI_DIR):
        logging.info("Creating directory '{}'".format(SOCO_CLI_DIR))
        try:
            os.mkdir(SOCO_CLI_DIR)
        except:
            error_report("Failed to create directory '{}'".format(SOCO_CLI_DIR))
            return

    logging.info("Saving shell history file: {}".format(HIST_FILE))
    try:
        readline.write_history_file(HIST_FILE)
    except Exception as e:
        logging.info("Error saving shell history file: {}".format(e))


def get_readline_history():
    if not os.path.exists(HIST_FILE):
        logging.info("No shell history file found: {}".format(HIST_FILE))
        return

    logging.info("Reading shell history file: {}".format(HIST_FILE))
    try:
        readline.read_history_file(HIST_FILE)
        readline.set_history_length(HIST_LEN)
    except Exception as e:
        logging.info("Error reading shell history file: {}".format(e))


def pretty_print_values(items, indent=2, separator=":", spacing=3, sort_by_key=False):
    """Print a list of keys and values.

    Args:
        items (dict): The keys and values to print.
        indent (int): Number of spaces to use as initial indentation.
        separator (str): The separator character(s) between the key and value.
        spacing (int): The minimum gap between the separator and
            the value.
        sort_by_key (bool): Whether to sort by item key.

    Example:
        One:     First
        Two:     Second
        Three:   Third
    """
    if len(items) == 0:
        return
    longest = max(len(key) for key in items)
    prefix = " " * indent
    key_vals = items.items()
    if sort_by_key:
        key_vals = sorted(key_vals)
    for key, value in key_vals:
        spacer = " " * (spacing + longest - len(key))
        print("{}{}{}{}{}".format(prefix, key, separator, spacer, value))


def playback_state(state):
    """Generate user-friendly playback states.

    Args:
        state (str): The Sonos-supplied state string.

    Returns:
        str: A user-friendly playback state description.
    """
    playback_mapping = {
        "STOPPED": "stopped",
        "PAUSED_PLAYBACK": "paused",
        "PLAYING": "in progress",
        "TRANSITIONING": "in a transitioning state",
    }
    try:
        return playback_mapping[state]
    except KeyError:
        return "unknown"


# Ensure that event subscriptions are cleared on CTRL-C
SUBS_LIST = set()


def remember_event_sub(sub):
    global SUBS_LIST
    logging.info("Adding event subscription record: '{}'".format(sub))
    SUBS_LIST.add(sub)


def forget_event_sub(sub):
    global SUBS_LIST
    try:
        logging.info("Removing event subscription record: '{}'".format(sub))
        SUBS_LIST.remove(sub)
    except KeyError:
        pass


def unsub_all_remembered_event_subs():
    global SUBS_LIST
    for sub in SUBS_LIST:
        try:
            event_unsubscribe(sub)
        except:
            break
    SUBS_LIST.clear()



"""Captures single keystrokes."""

import sys
from os import name as os_name

if os_name == "nt":
    import msvcrt

else:
    try:
        import termios

    except ImportError:
        pass


def get_keystroke() -> str:
    """Wait for a keypress, then return it."""

    # Windows
    if os_name == "nt":
        result = msvcrt.getch().decode()  # type: ignore

    # Unix
    else:
        result = None
        fd = sys.stdin.fileno()
        oldterm = termios.tcgetattr(fd)
        newattr = termios.tcgetattr(fd)
        newattr[3] = newattr[3] & ~termios.ICANON & ~termios.ECHO
        termios.tcsetattr(fd, termios.TCSANOW, newattr)
        try:
            result = sys.stdin.read(1)
        except IOError:
            pass
        finally:
            termios.tcsetattr(fd, termios.TCSAFLUSH, oldterm)

    return result



"""Matches a supplied speaker name to a stored name."""

import logging


def speaker_name_matches(name_supplied, name_stored):
    """Matches speaker names.

    Returns:
        bool, bool: Returns whether a match is found, and whether
            the match is considered exact or not.
    """

    # Exact match
    name_supplied_original = name_supplied
    name_stored_original = name_stored
    if name_supplied == name_stored:
        logging.info(
            "Found exact speaker name match for '{}'".format(name_stored_original)
        )
        return True, True

    # Case insensitive match; treat as an exact match
    name_supplied = name_supplied.lower()
    name_stored = name_stored.lower()
    if name_supplied == name_stored:
        logging.info(
            "Found case-insensitive exact speaker name match for '{}' as '{}'".format(
                name_supplied_original, name_stored_original
            )
        )
        return True, True

    # Normalised apostrophe match; treat as an exact match
    name_supplied = name_supplied.replace("’", "'")
    name_stored = name_stored.replace("’", "'")
    if name_supplied == name_stored:
        logging.info(
            "Found apostrophe-normalised exact speaker name match for '{}' as '{}'".format(
                name_supplied_original, name_stored_original
            )
        )
        return True, True

    # Partial match with start of name
    if name_stored.startswith(name_supplied):
        logging.info(
            "Found partial, start-of-name match for '{}' as '{}'".format(
                name_supplied_original, name_stored_original
            )
        )
        return True, False

    # Partial match with any part of name
    if name_supplied in name_stored:
        logging.info(
            "Found partial, any-part-of-name match for '{}' as '{}'".format(
                name_supplied_original, name_stored_original
            )
        )
        return True, False

    # Not found
    return False, False



import logging
import re
from datetime import datetime, timezone

from soco import SoCo  # type: ignore

from soco_cli.api import run_command


def track_follow(
    speaker: SoCo,
    use_local_speaker_list: bool = False,
    break_on_pause: bool = True,
    compact: bool = False,
):
    """Print out the 'track' details each time the track changes.

    Args:
        speaker (SoCo): The speaker to follow.
        use_local_speaker_list (bool, optional): Use cached discovery.
        break_on_pause (bool, optional): Whether to return control if the
            speaker enters the paused or stopped playback states.
        compact (bool, optional): Whether to use 'compact' output mode.

    This function operates as if 'outside' the main program logic, because
    it needs to output intermediate results as it executes. Hence, the
    'run_command()' API call is used.
    """

    def timestamp(short=False):
        local_tz = datetime.now(timezone.utc).astimezone().tzinfo
        if not short:
            return datetime.now(tz=local_tz).strftime("%H:%M:%S (%Z) %d-%b-%Y")
        else:
            return datetime.now(tz=local_tz).strftime("%H:%M")

    counter = 1
    print()
    while True:
        # If stopped, wait for the speaker to start playback
        _, state, _ = run_command(
            speaker, "state", use_local_speaker_list=use_local_speaker_list
        )
        if state in [
            "STOPPED",
            "PAUSED_PLAYBACK",
        ]:
            if not compact:
                print(
                    " [{}] Playback is stopped or paused at {}\n".format(
                        speaker.player_name, timestamp()
                    ),
                )
            else:
                print(
                    "{:5d}: [{}] Playback is stopped or paused".format(
                        counter, timestamp(short=True)
                    )
                )
                counter += 1
            if break_on_pause:
                logging.info("Playback is paused/stopped; returning")
                break
            logging.info("Playback is paused/stopped; waiting for start")
            run_command(
                speaker, "wait_start", use_local_speaker_list=use_local_speaker_list
            )
            logging.info("Speaker has started playback")

        # Print the track info
        exit_code, output, error_msg = run_command(
            speaker, "track", use_local_speaker_list=use_local_speaker_list
        )
        if exit_code == 0:
            # Manipulate output
            if "Using Line In" in output:
                line_in = True
                output = "   Playing from Line In\n"
            else:
                line_in = False
                output = output.split("\n ", 1)[1]
            if not compact:
                # Remove some of the line entries
                output = re.sub(".*Playback.*\\n", "", output)
                # output = re.sub(".*Elapsed.*\\n", "", output)
                output = re.sub(".*URI.*\\n", "", output)
                output = re.sub(".*Uri.*\\n", "", output)
                # Prefix speaker name and timestamp
                output = (
                    " [{}] Playing at ".format(speaker.player_name)
                    + timestamp()
                    + ":\n"
                    + output
                )
            else:  # Compact (one line) output
                if line_in:
                    output = "{:5d}: [{}] Playing from Line In".format(
                        counter, timestamp(short=True)
                    )
                else:
                    # Ordering of keys determines output order
                    keys = [
                        "Channel:",
                        "Radio Show:",
                        "Artist:",
                        "Creator(s):",
                        "Book Title:",
                        "Chapter:",
                        "Album:",
                        "Podcast:",
                        "Title:",
                        "Episode:",
                        "Release Date:",
                        "Narrator(s):",
                    ]
                    elements = {}
                    for line in output.splitlines():
                        for key in keys:
                            if key in line:
                                elements[key] = line.replace(key, "").lstrip()
                    output = "{:5d}: [{}] ".format(counter, timestamp(short=True))

                    # Prune fields for audio books
                    if "Book Title:" in elements:
                        elements.pop("Title:", None)
                        elements.pop("Narrator(s):", None)
                    first = True
                    for key in keys:
                        value = elements.pop(key, None)
                        if value:
                            if not first:
                                output = output + "| "
                            else:
                                first = False
                            output = output + key + " " + value + " "
            print(output)
        else:
            error_out = "{:5d}: [{}] {}".format(
                counter, timestamp(short=True), error_msg
            )
            print(error_out)

        logging.info("Waiting for end of track")
        run_command(
            speaker, "wait_end_track", use_local_speaker_list=use_local_speaker_list
        )
        counter += 1
